//			collapseAlign.C
//
// written by Zheng Gengbin

/*
This pass collapse an align tree constructed in prevoius pass in info.C 
to a one level tree . Deal with the condition when an array is align directly 
or indirectly with itself . Also when  *align_target, pass it down to its sons.
After this pass , we can assume any align is directly align to a different 
array (template).
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "pass.h"
#include "ext_func.h"


//                           align_1
//                            |
//                           align_2     (stmt2)
//                            |
//                           align_3     (stmt1)  about to change
void collapse_2_align(SgSymbol *align_1 ,SgSymbol *align_2 ,SgSymbol *align_3)
{
	SgStatement *stmt1,*stmt2;
	SgExpression *align_op,*align_dummy2,*align_with2,*align_dummy1,*align_with1;
	SgExpression *align_with_op1,*align_with_op2;
	SgSymbol     *align_target2;
	SgExpression *dummys , *exps,*with1;
	SgSymbol     *dumsym;

	stmt2=(SgStatement *)align_2->attributeValue(0,Align_Stmt);
	stmt1=(SgStatement *)align_3->attributeValue(0,Align_Stmt);

	align_op = stmt2->expr(2)->lhs();
        align_dummy2=align_op->lhs();
	align_with_op2=align_op->rhs();
        align_with2=align_op->rhs()->rhs()->lhs();
	align_target2 = align_op->rhs()->rhs()->symbol();

	align_op = stmt1->expr(2)->lhs();
        align_dummy1=align_op->lhs();
	align_with_op1=align_op->rhs();
        align_with1=align_op->rhs()->rhs()->lhs();

	for ( dummys=align_dummy2,exps=align_with1; 
		dummys ; dummys=dummys->rhs(),exps=exps->rhs() )
	{
		dummys->lhs()->symbol()->addAttribute(Dummy_Exp , exps->lhs() , sizeof(exps) );
	}
	
	SgExprListExp  *expr_list=NULL;
	for (exps = align_with2 ; exps ; exps=exps->rhs() )
	{
		// search align_dummy1 in this exps
		SgExpression *expr=exps->lhs();
		SgExpression *new_expr = expr->copyPtr();
		for (dummys=align_dummy2 ; dummys ; dummys=dummys->rhs() )
		{
			dumsym=dummys->lhs()->symbol();
			SgExpression *node=new_expr->IsSymbolInExpression(*dumsym); 
			if (node == NULL) continue;
			SgExpression *dum_exp=(SgExpression *)dumsym->attributeValue(0,Dummy_Exp);
			node->replace(*(dum_exp->copyPtr()));
		}
		SgExpression *exprlist=new SgExpression(EXPR_LIST,new_expr,NULL,NULL);
		if (expr_list == NULL)
			expr_list=(SgExprListExp *)exprlist;
		else
			expr_list->append(*exprlist);
	}
	align_op = stmt1->expr(2)->lhs();
        align_op->rhs()->rhs()->setLhs(expr_list);

	// change align_target
	align_op = stmt1->expr(2)->lhs();
	align_op->rhs()->rhs()->setSymbol(align_target2);

	// change the attributes
	align_2->deleteAttribute(0,Align_Son);
	align_3->deleteAttribute(0,Align_Parent);
	
	align_1->addAttribute(Align_Son, align_3 ,sizeof(align_3) );
	align_3->addAttribute(Align_Parent, align_1 ,sizeof(align_1) );

	// deal the * pass down the tree

	if (align_2->numberOfAttributes(Align_Star) > 0 )
		align_3->addAttribute(Align_Star,NULL,0);
	if (align_with_op2->lhs() )
		align_with_op1->setLhs(align_with_op2->lhs()->copy());

	// clear the attributes
	for ( dummys=align_dummy2,exps=align_with1; 
		dummys ; dummys=dummys->rhs(),exps=exps->rhs() )
	{
		dummys->lhs()->symbol()->deleteAttribute(0,Dummy_Exp);
	}
}

void collapse_aligns(SgStatement *hedr_stmt)
{
	SgSymbol * symbol = hedr_stmt->symbol()->next();
	SgSymbol * sym_son , *sym_par;

	while ((symbol!=NULL)) {
    		if (symbol->scope()==hedr_stmt)
      			break;
    		symbol=symbol->next();
  	}
	if (symbol==NULL) return ;

	int do_collapse ;
	while (1)
	{
	  int do_collapse = 0;
	  for( SgSymbol *symb=symbol ; symb; symb=symb->nextInScope()) 
	  {
		if (symb->scope() == NULL ) continue;
//printf("checking %s :-- %d : %d\n",symb->identifier(),symb->numberOfAttributes(Align_Parent),symb->numberOfAttributes(Align_Son));
		if (symb->numberOfAttributes(Align_Parent) > 0 &&
			symb->numberOfAttributes(Align_Son) >0 )
		{
			sym_son=(SgSymbol *)symb->attributeValue(0,Align_Son);
			sym_par=(SgSymbol *)symb->attributeValue(0,Align_Parent);
			  // Allocatable , remain
	    		if (sym_son->attributes() &  ALLOCATABLE_BIT) continue;
			  // if is collapse to itself
			if (sym_son == sym_par) continue;
printf("Collapse %s -- %s -- %s \n",sym_par->identifier(), symb->identifier(),sym_son->identifier() );
			collapse_2_align(sym_par,symb,sym_son);
			do_collapse = 1;
		}
	  }
	  // if nothing is done
	  if (do_collapse == 0) break;
	}
}

